rawset(_G, "PetList", {}) //the mighty table of pets
rawset(_G, "PetAbilities", false) //defence/offense enabled

freeslot(
"MT_PET",
"S_PET"
 )

states[S_PET] = {SPR_NULL, A, -1, nil, 0, 0, S_PET}

local MAXPETSPRITES = 10 //modify me if you want more/less max amount of sprites usable for all pets animations!
local PETSKINNAMELIMIT = 15 //modify me to change the max length of a pet skin's name

//pet skin adding
rawset(_G, "Pet_AddPet", function(btable, name)
	local ERROR = "\133ERROR: \130"
	
	if not (btable)
	or not (name) then
		print(ERROR+"NO PET SKIN INFO/NAME GIVEN FOR A SKIN, CHECK YOUR CODE!")
		return
	end
	
	//rename all entries to lowercase for ease
	local newtable = {}
	for n,e in pairs(btable)
		newtable[string.lower(n)] = e

		//for the tables in this table
		if (type(e) == "table") then
			local cur = newtable[string.lower(n)]
			local mini = {}
			for mn, me in pairs(e)
				mini[string.lower(mn)] = me
			end
			newtable[string.lower(n)] = mini
		end
	end
	
	local ptable = newtable

	
	//lowercase only bud
	name = string.lower($1)
	
	if (name == "none") then
		print(ERROR+"none IS A RESERVED PET SKIN NAME")
		return
	end
	
	//keep added pet skin names within the defined limits
	if (string.len(name) > PETSKINNAMELIMIT) then
		name = string.sub($1, 1, PETSKINNAMELIMIT)
		print(ERROR+"GIVEN SKIN NAME IS TOO LONG, LIMIT IS "+PETSKINNAMELIMIT+", CUT OFF TO: "+name)
	end
	
	//face the wrath of the pet judges
	if not (ptable.spriteset) then
		print(ERROR+"NO SPRITESET GIVEN FOR PET SKIN: "+name+"\nSKIN NOT ADDED")
		return
	end

	if (ptable.anims) then
		//go through each table in the table
		for i,_ in pairs(ptable.anims)
		
			//somehow not a table in here
			if not (type(ptable.anims[i]) == "table") then
				print(ERROR+"NON-TABLE ENTRY IN ANIMATIONS FOR "+name)
				continue
			end
			
			//go through the 
			for x,e in pairs(ptable.anims[i])
			
				//we only want the first two entries
				if (x >= 3) then continue end
				if (type(e) == "string") then
					ptable.anims[i][x] = R_Char2Frame(e)
				end
			end
		end
	end
	
	//color checking if applicable
	if (ptable.color) then
		if (type(ptable.color) == "string") then
			ptable.color = R_GetColorByName($1)
		end
		
		if not (type(ptable.color) == "number") then
			print(ERROR+"INVALID COLOR GIVEN FOR PET SKIN: "+name+"\nDEFAULTING TO NONE")
			ptable.color = nil
		end
	end
	
	PetList[name] = ptable
	print("\134"+"ADDED PET SKIN: \128"+name)
end)

local function PetSave(player)
	if not (player == consoleplayer) then return end
	//write & update, clears the file before writing
	local file = io.openlocal("client/PetSettings.dat", "w+")

	file:write("skin "+player.petskin)
	file:write("\ncolor "+player.petcolor)
	file:write("\nheight "+player.petheight+"\n")

	io.close(file)
end

local function UpdatePetAnim(mo, anim, skin)

	//don't keep settin' it, loser
	if (mo.animset == anim)
	or not (skin)
	or (skin == "none")
	or not (PetList[skin].anims[anim]) then return end

	skin = PetList[$1]

	mo.frame = skin.anims[anim][1]
	mo.animset = anim
	mo.petanimdur = skin.anims[anim][3]
end

//Fake FF_ANIMATE so I can do my own funky stuff with it!
local function PetAnimate(mo, skin)

	skin = PetList[$1]

	if not (mo.animset) 
	or not (skin)
	or not (skin.anims) then return end

	if not (mo.petanimdur)
	or (mo.petanimdur == 0) then

		if (mo.frame+1 <= skin.anims[mo.animset][2])
			mo.frame = $1+1
		else
			mo.frame = skin.anims[mo.animset][1]
		end
		mo.petanimdur = skin.anims[mo.animset][3]
	else
		mo.petanimdur = $1-1
	end
	
end

//Pet Skin Changing
local function UpdatePetSkin(pet, nocolor)
	//how'd you get here?
	if not (pet.target)
	or not (pet.target.player) then return end
	
	//invalid skin, reset pet
	if not (pet.target.player.petskin)
	or not (PetList[pet.target.player.petskin]) then
		pet.state = S_PET
		pet.sprite = SPR_NULL
		pet.shadowscale = 0
		pet.animset = "idle"
		return
	end

	local PetSkin = PetList[pet.target.player.petskin]

	
	if (PetSkin.anims) then
		UpdatePetAnim(pet, "idle", pet.target.player.petskin)
	end
	
	//don't update colors if unwanted
	if not (nocolor) then
		if (PetSkin.color) then
			pet.color = PetSkin.color
		else
			pet.color = SKINCOLOR_NONE
		end
		pet.target.player.petcolor = pet.color
	end
	
	pet.shadowscale = FRACUNIT/6
	
	if (PetSkin.shadowscale) then
		pet.shadowscale = PetSkin.shadowscale
	end
	
	pet.frame = A
	pet.sprite = PetSkin.spriteset
end


///
///		PET COMMANDS
///

//changin' pet skins
local function ChangePetSkin(player, skin)
	if not (skin)
	or not (PetList[string.lower(skin)]) then
		if not (skin == "none") then
			CONS_Printf(player, "\130No/invalid pet skin given, to remove your pet use the skin \131none\130, to view all pets try the command \131pet list")
			return
		else
			if (player.mo)
			and (player.mo.pet) then
				player.petskin = nil
				UpdatePetSkin(player.mo.pet)
			end
		end
	end

	if (string.len(skin) > PETSKINNAMELIMIT) then
		CONS_Printf(player, "\130Given pet skin name is too long")
	end
	//lowercase club!!
	player.petskin = string.lower(skin)
	if (player.mo)
	and (player.mo.pet) then
		UpdatePetSkin(player.mo.pet)
	end
	PetSave(player)
end


//the pet list, it lists the pets
local function ListPets(player)
	CONS_Printf(player, "\130Available pets:")
	for i,_ in pairs(PetList)
		CONS_Printf(player, "\131"+i)
	end
	CONS_Printf(player, "\130Press PageUp/PageDown to scroll the console to view the full list if needed")
end


//recolor that pet into your own oc!!
local function ColorPet(player, color)

	if not (player.mo)
	or not (player.mo.valid) then
		print("\130You must be in a level to use this")
		return
	end

	if not (player.mo.pet.color == SKINCOLOR_NONE) then

		//let's just emulate how it works for player colors
		if not (color) then
			CONS_Printf(player, "\130Current pet color is "+R_GetNameByColor(player.mo.pet.color)+" default for pet is "+R_GetNameByColor(PetList[player.petskin].color))
			return 
		end

		//for the wacky folk who use color numbers
		if (tonumber(color)) then
			color = tonumber($1)
		end

		if (type(color) == "string") then
			local oldcolor = color
			color = R_GetColorByName(string.upper($1))
			
			//prevent invalid colors setting them to green
			if (color == SKINCOLOR_GREEN)
			and not (string.lower(oldcolor) == "green") then
				CONS_Printf(player, "\130"+oldcolor+" is not a valid color")
				return
			end
		end
		
		//check it's within range
		if (color > #skincolors-1)
		or (color < 1) then
			CONS_Printf(player, "\130"+color+" is not a valid color")
			return
		end

		player.petcolor = color
		player.mo.pet.color = color
		PetSave(player)
	else
		CONS_Printf(player, "\130This pet cannot be recolored")
	end
end

//get outta the way i'm watchin' tv!
local function SetPetHeight(player, heightval)

	if not (tonumber(heightval))
	and not (tonumber(heightval) == 0) then
		CONS_Printf(player, "\130"+heightval+" is not a valid number, decimals are not supported.")
		return
	end

	heightval = tonumber($1)
	
	if (heightval > 30) then
		heightval = 30
	elseif (heightval < -30) then
		heightval = -30
	end
	
	player.petheight = heightval
	PetSave(player)
end


//helpin' hand
local function PetCommandsHelp(player, command)
	
	//all hail tables
	local helptext = {
	help = "\130List of commands:\n\131skin\nlist\ncolor\nhelp\nheight\nguard",
	skin = "\131Pet Skin:\130 sets your pet to the given pet skin, use \131pet list\130 to view available skins",
	list = "\131Pet List:\130 lists the pet skins available to use with \131pet skin",
	color = "\131Pet Color:\130 sets your pet to the given color if the pet's color can be changed, numbers and color names are both usable",
	colour = "\131Pet Colour:\130 sets your pet to the given colour if the pet's colour can be changed, numbers and colour names are both usable",
	height = "\131Pet Height:\130 offsets your pet's height that it hovers at by the given number, ranges from -30 to 30",
	guard = "\131Pet Guard:\130 turns the pet defense ability on or off, admin only"
	}
	//list off the commands instead
	if not (command) then
		CONS_Printf(player, helptext.help)
		CONS_Printf(player, "\130For help on a specific comamnd enter \131pet help commandname")
		return
	end
	
	if not (helptext[command]) then
		CONS_Printf(player, "\130Unknown command, for a list of commands enter \131pet help") //GET OUTTA THERE
	else
		CONS_Printf(player, helptext[command])
	end

end

local function PetToggleAbilities(player, var)
	if (netgame)
	and not (IsPlayerAdmin(player))
	and not (player == server) then
		CONS_Printf(player, "\130This command is admin only!")
		return
	end
	
	if not (var) then
		CONS_Printf(player, "\130Pet Guard is currently "+PetAbilities)
		return
	end
	
	if (var == "on") then
		CONS_Printf(player, "\130Pet Guard enabled")
		PetAbilities = true
		return
	end
	
	if (var == "off") then
		CONS_Printf(player, "\130Pet Guard disabled")
		PetAbilities = false
		return
	end
	
	CONS_Printf(player, "\130To toggle Pet Guard, please use on/off")
end


local function PetCommand(player, var1, var2)
	
	local commands = {
	help = PetCommandsHelp,
	skin = ChangePetSkin,
	list = ListPets,
	color = ColorPet,
	colour = ColorPet,
	height = SetPetHeight,
	guard = PetToggleAbilities
	}
	
	
	if not (var1)
	or not (commands[var1]) then
		CONS_Printf(player, "\130For a list of commands, enter \131pet help")
		return
	end

	//safety for YELLING and ThIs
	var1 = string.lower($1)

	if (var2) then
		var2 = string.lower($1)
	end
	commands[var1](player, var2)
end

COM_AddCommand("pet", PetCommand)


///
///		PET IO
///

local function PetLoad(player)
	if not (player == consoleplayer) then return end
	local file = io.openlocal("client/PetSettings.dat", "r")

	if not (file) then
		return false
	end

	
	for line in file:lines() do
		//space denotes the end of the entry name
		local nameend = string.find(line, " ")
		if not (nameend) then continue end
		//everything before the space is the entry name
		local varname = string.sub(line, 0, nameend-1)
		//everything after it is the value
		local varval = string.sub(line, nameend+1)

		PetCommand(player, varname, varval)
	end
	io.close(file)
end



///
///		MAIN PET STUFF
///

 mobjinfo[MT_PET] = {
	spawnstate = S_PET,
	deathstate = S_PET,	//death can't save you now
	radius = 20*FRACUNIT,
	height = 10*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOBLOCKMAP
}


local function MoveLag(mo, domove)

	local nangle = R_PointToAngle2(mo.target.x, mo.target.y, mo.x, mo.y)
	local posangle = mo.target.player.drawangle+ANGLE_180+ANG20
	local radius = 46*FRACUNIT
	local newz
	
	//gravity flipping stuff!!
	if (mo.target.eflags & MFE_VERTICALFLIP) then
		newz = mo.target.z - mo.target.height
		mo.eflags = $1|MFE_VERTICALFLIP
	else
		newz = mo.target.z + mo.target.height
		mo.eflags = $1 & ~MFE_VERTICALFLIP
	end
	
	
	local newpos = {
	x = mo.target.x + FixedMul(cos(posangle), radius),
	y = mo.target.y + FixedMul(sin(posangle), radius),
	z = newz,
	angle = mo.target.player.drawangle
	}
	
	if (mo.target.player.powers[pw_carry] == CR_NIGHTSMODE)
		newpos.angle = mo.target.player.flyangle
	end
	
	
	if not (mo.lagtable) then
		mo.lagtable = {
		newpos,
		newpos,
		newpos
		}
	end

	local SPEED = 4
	local HEIGHT = 10
	local DISTANCE = 6
	
	local newx = mo.lagtable[1].x 
	local newy = mo.lagtable[1].y 
	local newz = mo.lagtable[1].z
	//update the table, without moving the pet
	if (domove) then
		mo.angle = mo.lagtable[1].angle
		P_TeleportMove(mo, newx, newy, newz+(mo.target.player.petheight*mo.target.scale))
	end
	
	table.remove(mo.lagtable, 1)
	table.insert(mo.lagtable, newpos)
	
end


local function PetFollow(mo)
	
	//no buddy
	if not (mo.target) then return end
	local offset = 20

	//got a pet already?
	if (mo.target.player)
	and (mo.target.player.chipactive) then
		offset = -20
	end
	
	if (mo.target.player.petskin) then
		PetAnimate(mo, mo.target.player.petskin)
	end
	

	//if you're not dead you get to move
	if (mo.target.player.playerstate == PST_LIVE) then
		
		local domove = true
		
		if (mo.flags & MF_MISSILE)
		or (mo.target.player.powers[pw_carry] == CR_NIGHTSMODE) then
			local x,y,z
			local radius = 32*FRACUNIT
			local angle = mo.target.petang
			
			x = mo.target.x + FixedMul(cos(angle), radius)
			y = mo.target.y + FixedMul(sin(angle), radius)
			z = mo.target.z+mo.target.height/5
			mo.angle = angle+ANGLE_90
			
			P_TeleportMove(mo, x, y, z)
			domove = false
		end
		 
		MoveLag(mo, domove)
	end
end

addHook("MobjThinker", PetFollow, MT_PET)

//checks for collision when orbiting since MF_MISSILE doesn't suit our needs
local function PetCollide(pet, mo)	
	if not (pet.flags & MF_MISSILE)
	or (pet.target == mo)
	or not (mo.flags & MF_SHOOTABLE)
		return
	end
	
	//don't attack monitors
	if not (mo.flags & MF_ENEMY)
	and not (mo.flags & MF_BOSS)
	or (mo.type == MT_METALSONIC_RACE) then
		return false
	end

	if (pet.z > mo.z + mo.height) //above
	or (pet.z + pet.height < mo.z) then //below
		return
	end
	
	pet.flags = mobjinfo[pet.type].flags
	P_DamageMobj(mo, pet, pet.target)
end
addHook("MobjMoveCollide", PetCollide, MT_PET)



//metal sonic kills your pet, let's fix that
local function PetDeath(mo)
	UpdatePetSkin(mo)
	return true //i refuse
end
addHook("MobjDeath", PetDeath, MT_PET)

local function PetThinkFrame()
	for player in players.iterate
		if not player.mo then return end
		
		local AnimStrings = {
		"idle",		//PA_ETC
		"idle",		//PA_IDLE
		"idle",		//PA_EDGE
		"move",		//PA_WALK
		"move",		//PA_RUN
		"move",		//PA_DASH
		"pain",		//PA_PAIN
		"move",		//PA_ROLL
		"jump",		//PA_JUMP
		"jump",		//PA_SPRING
		"fall",		//PA_FALL
		"move",		//PA_ABILITY - abilities
		"move"		//PA_ABILITY2 - spindash/fang popgun
		}
	
		if (player.playerstate == PST_LIVE) then
			//no pet, set up pet variables
			if not (player.mo.pet) then
				local docolorupdate = true
				player.pet = P_SpawnMobj(player.mo.y, player.mo.y, player.mo.z + player.mo.height, MT_PET)
				//pet removed itself immediately
				if not (player.pet.valid) then
					player.mo.pet = nil
					player.pet = nil
					return
				end
				player.mo.pet = player.pet
				player.mo.pet.target = player.mo
				player.mo.petang = 0

				PetLoad(player)
				
				if (player.petcolor) then
					docolorupdate = false
				end
				
				UpdatePetSkin(player.mo.pet, true)
				
				if not (player.petheight) then
					player.petheight = 0
				end
				
			end
			
			if (player.pet.startorbit)
			and (P_IsObjectOnGround(player.mo)) then
				player.pet.startorbit = nil
				player.pet.flags = MF_MISSILE|MF_NOGRAVITY|MF_NOCLIPHEIGHT
			end
			
			if (player.petskin)
			and not (player.petskin == "none") then
				
				//update pet animations if there's an appropriate one provided
				local PlayAnim = "idle"
				if (AnimStrings[player.panim+1])
				and (PetList[player.petskin].anims[AnimStrings[player.panim+1]])
				and not (player.mo.pet.flags & MF_MISSILE)
				and not (player.powers[pw_carry] == CR_NIGHTSMODE) then
					PlayAnim = AnimStrings[player.panim+1]
				end

				if (player.mo.pet.flags & MF_MISSILE)
				or (player.powers[pw_carry] == CR_NIGHTSMODE)
					if (PetList[player.petskin].anims["move"]) then
						PlayAnim = "move" //move it if you can when orbitin'
					end
				end
				
				UpdatePetAnim(player.mo.pet, PlayAnim, player.petskin)
				
			end
			
			local SPEED = 4
			local HEIGHT = 10
			local DISTANCE = 6
			P_TeleportMove(player.mo.pet, player.mo.pet.x, player.mo.pet.y, HEIGHT*FRACUNIT+player.mo.pet.z+DISTANCE*sin(player.mo.petang))
			player.mo.petang = $1+ANG1*SPEED
		else
			//dead
			if (player.mo.pet) then
				UpdatePetAnim(player.mo.pet, "dead", player.petskin)
			end
		end
	end
end

addHook("ThinkFrame", PetThinkFrame)

local function PetHurtCheck(mo)
	if not (mo.pet)
	or not (PetAbilities) then
		return
	end
	
	if not (mo.pet.startorbit) then
		mo.pet.startorbit = true
	end
end

addHook("MobjDamage", PetHurtCheck, MT_PLAYER)

//Removes pet after death
local function RemovePet(mo)
	if not (multiplayer) then return end
	if (mo.pet)
	and (mo.pet.valid)
		P_RemoveMobj(mo.pet)
		mo.pet = nil
	end
end
addHook("MobjRemoved", RemovePet, MT_PLAYER)